/*
 * WsDmaPort.c
 *
 * Created: 10.10.2018 18:45:49
 *  Author: Thomas F
 */ 

#include <stdbool.h>
#include <string.h>
#include "sam.h"

#include "clocks.h"
#include "misc.h"
#include "IO.h"
#include "dma.h"
#include "WsDmaPort.h"


extern DmacDescriptors_t WsDmacDescriptors;

static _Bool		InitOk;
static uint8_t		UsedPinPattern;
static uint8_t*		DmaBuffer;
static uint8_t**	RgbData;
static uint16_t		LedCount;

static uint8_t		PortPatternSetClrHigh;
static uint32_t		DmaBufferSize;


// ----------------------------------------------------------------------------------------------


int WsDmaPortInit(PortGroup* portGroup,  uint8_t portByteOffset, uint8_t usedPinPattern, uint8_t* dmaBuffer, uint16_t dmaBufferSize, uint8_t* rgbData[], _Bool signalInvert)
{
	
	InitOk = false;

	
	// ##### Check global DMA settings (DMA must be enabled and configured in the right way)
	
	if (DMAC->PRICTRL0.bit.RRLVLEN3)	// Round-Robin must be off
		return -1;
	else if (DMAC->PRICTRL0.bit.QOS3 != DMAC_PRICTRL0_QOS3_CRITICAL_Val)	// QoS must be "Latency critical"
		return -2;
	else if (!DMAC->CTRL.bit.LVLEN3)	// Priority Level 3 must be enabled	
		return -3;
	else if (!DMAC->CTRL.bit.DMAENABLE)	// Last but not least the DMA must be enabled
		return -4;
	
	
	// ##### Check parameters
	
	if (dmaBuffer == 0)				// A DMA buffer would be nice :-)
		return -10;
	else if (((uint32_t)dmaBuffer & 0x03) != 0)	// Word alignment is necessary
		return -11;
	else if (dmaBufferSize < 24)	// Memory for at least one LED per channel
		return -12;
	else if (portByteOffset > 3)	// A word is 4 bytes! (ARM)
		return -13;
	else if (usedPinPattern == 0)	// All this work for nothing?
		return -14;
	

	// ##### Save parameters
	
	UsedPinPattern	= usedPinPattern;
	DmaBuffer		= dmaBuffer;
	RgbData			= rgbData;
	DmaBufferSize	= dmaBufferSize;
	LedCount		= dmaBufferSize / 24;

	PortPatternSetClrHigh = usedPinPattern;


	// ##### Port
	
	// Direction must be set separately (For safety reasons) 
	if (signalInvert)
	{
		*((uint8_t*)&portGroup->OUTSET.reg + portByteOffset) = usedPinPattern;
	}
	else
	{
		*((uint8_t*)&portGroup->OUTCLR.reg + portByteOffset) = usedPinPattern;
	}
	

	// ##### DMA 

	uint32_t regOutSet = (uint32_t)&portGroup->OUTSET.reg + portByteOffset;
	uint32_t regOutClr = (uint32_t)&portGroup->OUTCLR.reg + portByteOffset;



	// Channel 0 Port set
	DMAC->Channel[0].CHCTRLA.reg  = DMAC_CHCTRLA_TRIGSRC(TC4_DMAC_ID_OVF)
								  | DMAC_CHCTRLA_TRIGACT_BURST
								  | DMAC_CHCTRLA_BURSTLEN_SINGLE
								  | DMAC_CHCTRLA_THRESHOLD_1BEAT;
	
	DMAC->Channel[0].CHPRILVL.reg = DMAC_CHPRILVL_PRILVL_LVL3;
	DMAC->Channel[0].CHEVCTRL.reg = DMAC_CHEVCTRL_EVACT_NOACT;
	
	
	WsDmacDescriptors.C0D1.BTCTRL.reg = DMAC_BTCTRL_BEATSIZE_BYTE
									  | DMAC_BTCTRL_BLOCKACT_NOACT
									  | DMAC_BTCTRL_EVOSEL_DISABLE
									  | DMAC_BTCTRL_VALID;
	WsDmacDescriptors.C0D1.DESCADDR.reg = 0;
	WsDmacDescriptors.C0D1.BTCNT.reg	= DmaBufferSize;
	WsDmacDescriptors.C0D1.SRCADDR.reg	= (uint32_t)&PortPatternSetClrHigh;
	if (!signalInvert)
		WsDmacDescriptors.C0D1.DSTADDR.reg = regOutSet;
	else
		WsDmacDescriptors.C0D1.DSTADDR.reg = regOutClr;



	// Channel 1 Port clr low
	DMAC->Channel[1].CHCTRLA.reg  = DMAC_CHCTRLA_TRIGSRC(TC4_DMAC_ID_MC_0)
								  | DMAC_CHCTRLA_TRIGACT_BURST
								  | DMAC_CHCTRLA_BURSTLEN_SINGLE
								  | DMAC_CHCTRLA_THRESHOLD_1BEAT;
	
	DMAC->Channel[1].CHPRILVL.reg = DMAC_CHPRILVL_PRILVL_LVL3;
	DMAC->Channel[1].CHEVCTRL.reg = DMAC_CHEVCTRL_EVACT_NOACT;
	
	
	WsDmacDescriptors.C1D1.BTCTRL.reg = DMAC_BTCTRL_STEPSIZE_X1
									  | DMAC_BTCTRL_STEPSEL_SRC
									  | DMAC_BTCTRL_SRCINC
									  | DMAC_BTCTRL_BEATSIZE_BYTE
									  | DMAC_BTCTRL_BLOCKACT_NOACT
									  | DMAC_BTCTRL_EVOSEL_DISABLE
									  | DMAC_BTCTRL_VALID;
	WsDmacDescriptors.C1D1.DESCADDR.reg = 0;
	WsDmacDescriptors.C1D1.BTCNT.reg	= DmaBufferSize;
	WsDmacDescriptors.C1D1.SRCADDR.reg	= (uint32_t)DmaBuffer + DmaBufferSize;
	if (!signalInvert)
		WsDmacDescriptors.C1D1.DSTADDR.reg = regOutClr;
	else
		WsDmacDescriptors.C1D1.DSTADDR.reg = regOutSet;
	


	// Channel 2 Port clr high
	DMAC->Channel[2].CHCTRLA.reg  = DMAC_CHCTRLA_TRIGSRC(TC4_DMAC_ID_MC_1)
								  | DMAC_CHCTRLA_TRIGACT_BURST
								  | DMAC_CHCTRLA_BURSTLEN_SINGLE
								  | DMAC_CHCTRLA_THRESHOLD_1BEAT;
	
	DMAC->Channel[2].CHPRILVL.reg = DMAC_CHPRILVL_PRILVL_LVL3;
	DMAC->Channel[2].CHEVCTRL.reg = DMAC_CHEVCTRL_EVACT_NOACT;
	
	
	WsDmacDescriptors.C2D1.BTCTRL.reg = DMAC_BTCTRL_BEATSIZE_BYTE
									  | DMAC_BTCTRL_BLOCKACT_NOACT
									  | DMAC_BTCTRL_EVOSEL_DISABLE
									  | DMAC_BTCTRL_VALID;
	WsDmacDescriptors.C2D1.DESCADDR.reg = 0;
	WsDmacDescriptors.C2D1.BTCNT.reg	= DmaBufferSize;
	WsDmacDescriptors.C2D1.SRCADDR.reg	= (uint32_t)&PortPatternSetClrHigh;
	if (!signalInvert)
		WsDmacDescriptors.C2D1.DSTADDR.reg = regOutClr;
	else
		WsDmacDescriptors.C2D1.DSTADDR.reg = regOutSet;
			

	
	// ##### TC 

	// Enable module clock for TC
	MCLK->APBCMASK.bit.TC4_ = true;
	
	// Turn on clock for TC core
	GCLK->PCHCTRL[TC4_GCLK_ID].reg = GCLK_PCHCTRL_GEN(GEN_GCLK_120M)
								   | GCLK_PCHCTRL_CHEN;
	
	while(GCLK->PCHCTRL[TC4_GCLK_ID].bit.CHEN == false) { ; }


	// Initializing is done
	InitOk = true;
	
	
// For debugging purposes (low/high time on pin)
PORTB->PMUX[WS_LOW_A4_PB08_Pos/2].bit.PMUXE		= PMUX_E;
PORTB->PMUX[WS_HIGH_A5_PB09_Pos/2].bit.PMUXO	= PMUX_E;
	
PORTB->PINCFG[WS_LOW_A4_PB08_Pos].reg	= PORT_PINCFG_PMUXEN;
PORTB->PINCFG[WS_HIGH_A5_PB09_Pos].reg	= PORT_PINCFG_PMUXEN;


	return 0;
}


// ----------------------------------------------------------------------------------------------


int WsDmaPortStartTransfer(void)
{
	// Check state
	if (!InitOk)	// WsDmaPortInit() not called or error
		return -1;
	else if (IsWsDmaPortTransferActiv())	// Last transfer still active
		return -2;

	
	// Reset is necessary to clear active DMA triggers
	TC4->COUNT8.CTRLA.bit.SWRST = true;
	while(TC4->COUNT8.SYNCBUSY.bit.SWRST) { ; }
	
	// Set timer to 800000kHz and 8Bit mode
	TC4->COUNT8.CTRLA.reg = TC_CTRLA_MODE_COUNT8 | TC_CTRLA_PRESCALER_DIV1;

	#define TC_PER ((120000000 / 800000) - 1)	// 800 kHz

	TC4->COUNT8.PER.reg = TC_PER;
	while(TC4->COUNT8.SYNCBUSY.bit.PER) { ; }

	// Set counter that first DMA trigger is coming from channel0
	TC4->COUNT8.COUNT.reg = TC_PER;
	while(TC4->COUNT8.SYNCBUSY.bit.COUNT) { ; }
	
	// Set low and high Bit time
	TC4->COUNT8.CC[0].reg = (0.3/1.25 * (TC_PER + 1));
	TC4->COUNT8.CC[1].reg = (0.8/1.25 * (TC_PER + 1));
	while(TC4->COUNT8.SYNCBUSY.reg & (TC_SYNCBUSY_CC0 | TC_SYNCBUSY_CC1)) { ; }
	
// For debugging purposes (low/high time on pin)
TC4->COUNT8.WAVE.bit.WAVEGEN = TC_WAVE_WAVEGEN_NPWM_Val;
	
	// Enable DMA channels
	DMAC->Channel[0].CHCTRLA.bit.ENABLE = true;
	DMAC->Channel[1].CHCTRLA.bit.ENABLE = true;
	DMAC->Channel[2].CHCTRLA.bit.ENABLE = true;

	// Start counter
	TC4->COUNT8.CTRLA.bit.ENABLE = true;

	return 0;
}


// ----------------------------------------------------------------------------------------------


_Bool IsWsDmaPortTransferActiv(void)
{
	return (DMAC->Channel[2].CHCTRLA.bit.ENABLE == true);
}


// ----------------------------------------------------------------------------------------------


int ConvertRgbToDmaOneChannel(void)
{
	// Check state
	if (!InitOk)	// WsDmaPortInit() not called or error
		return -1;
	
	
	uint8_t* dmaBuffer = DmaBuffer;
	uint8_t* rgbData   = RgbData[0];
	
	for (int l = 0; l < LedCount; l++)
	{
		int cb;
		int i = 0;
		
		cb = ~rgbData[1];	// Green
		do
		{
			*dmaBuffer++ = (cb & 0x80) ? UsedPinPattern : 0;
			cb <<= 1;
		} while ( ++i < 8);
		
		cb = ~rgbData[0];	// Red
		do
		{
			*dmaBuffer++ = (cb & 0x80) ? UsedPinPattern : 0;
			cb <<= 1;
		} while ( ++i < 16);

		cb = ~rgbData[2];	// Blue
		do
		{
			*dmaBuffer++ = (cb & 0x80) ? UsedPinPattern : 0;
			cb <<= 1;
		} while ( ++i < 24);
		
		rgbData += 3;
	}
	
	return 0;
}


// ----------------------------------------------------------------------------------------------


#ifdef WS_RGB2DMA_PORT_FAST
	
static uint32_t BitSwapTab[512] =
{
	0x80808080, 0x80808080, 0x80808080, 0x00808080, 0x80808080, 0x80008080, 0x80808080, 0x00008080,
	0x80808080, 0x80800080, 0x80808080, 0x00800080, 0x80808080, 0x80000080, 0x80808080, 0x00000080,
	0x80808080, 0x80808000, 0x80808080, 0x00808000, 0x80808080, 0x80008000, 0x80808080, 0x00008000,
	0x80808080, 0x80800000, 0x80808080, 0x00800000, 0x80808080, 0x80000000, 0x80808080, 0x00000000,
	0x00808080, 0x80808080, 0x00808080, 0x00808080, 0x00808080, 0x80008080, 0x00808080, 0x00008080,
	0x00808080, 0x80800080, 0x00808080, 0x00800080, 0x00808080, 0x80000080, 0x00808080, 0x00000080,
	0x00808080, 0x80808000, 0x00808080, 0x00808000, 0x00808080, 0x80008000, 0x00808080, 0x00008000,
	0x00808080, 0x80800000, 0x00808080, 0x00800000, 0x00808080, 0x80000000, 0x00808080, 0x00000000,
	0x80008080, 0x80808080, 0x80008080, 0x00808080, 0x80008080, 0x80008080, 0x80008080, 0x00008080,
	0x80008080, 0x80800080, 0x80008080, 0x00800080, 0x80008080, 0x80000080, 0x80008080, 0x00000080,
	0x80008080, 0x80808000, 0x80008080, 0x00808000, 0x80008080, 0x80008000, 0x80008080, 0x00008000,
	0x80008080, 0x80800000, 0x80008080, 0x00800000, 0x80008080, 0x80000000, 0x80008080, 0x00000000,
	0x00008080, 0x80808080, 0x00008080, 0x00808080, 0x00008080, 0x80008080, 0x00008080, 0x00008080,
	0x00008080, 0x80800080, 0x00008080, 0x00800080, 0x00008080, 0x80000080, 0x00008080, 0x00000080,
	0x00008080, 0x80808000, 0x00008080, 0x00808000, 0x00008080, 0x80008000, 0x00008080, 0x00008000,
	0x00008080, 0x80800000, 0x00008080, 0x00800000, 0x00008080, 0x80000000, 0x00008080, 0x00000000,
	0x80800080, 0x80808080, 0x80800080, 0x00808080, 0x80800080, 0x80008080, 0x80800080, 0x00008080,
	0x80800080, 0x80800080, 0x80800080, 0x00800080, 0x80800080, 0x80000080, 0x80800080, 0x00000080,
	0x80800080, 0x80808000, 0x80800080, 0x00808000, 0x80800080, 0x80008000, 0x80800080, 0x00008000,
	0x80800080, 0x80800000, 0x80800080, 0x00800000, 0x80800080, 0x80000000, 0x80800080, 0x00000000,
	0x00800080, 0x80808080, 0x00800080, 0x00808080, 0x00800080, 0x80008080, 0x00800080, 0x00008080,
	0x00800080, 0x80800080, 0x00800080, 0x00800080, 0x00800080, 0x80000080, 0x00800080, 0x00000080,
	0x00800080, 0x80808000, 0x00800080, 0x00808000, 0x00800080, 0x80008000, 0x00800080, 0x00008000,
	0x00800080, 0x80800000, 0x00800080, 0x00800000, 0x00800080, 0x80000000, 0x00800080, 0x00000000,
	0x80000080, 0x80808080, 0x80000080, 0x00808080, 0x80000080, 0x80008080, 0x80000080, 0x00008080,
	0x80000080, 0x80800080, 0x80000080, 0x00800080, 0x80000080, 0x80000080, 0x80000080, 0x00000080,
	0x80000080, 0x80808000, 0x80000080, 0x00808000, 0x80000080, 0x80008000, 0x80000080, 0x00008000,
	0x80000080, 0x80800000, 0x80000080, 0x00800000, 0x80000080, 0x80000000, 0x80000080, 0x00000000,
	0x00000080, 0x80808080, 0x00000080, 0x00808080, 0x00000080, 0x80008080, 0x00000080, 0x00008080,
	0x00000080, 0x80800080, 0x00000080, 0x00800080, 0x00000080, 0x80000080, 0x00000080, 0x00000080,
	0x00000080, 0x80808000, 0x00000080, 0x00808000, 0x00000080, 0x80008000, 0x00000080, 0x00008000,
	0x00000080, 0x80800000, 0x00000080, 0x00800000, 0x00000080, 0x80000000, 0x00000080, 0x00000000,
	0x80808000, 0x80808080, 0x80808000, 0x00808080, 0x80808000, 0x80008080, 0x80808000, 0x00008080,
	0x80808000, 0x80800080, 0x80808000, 0x00800080, 0x80808000, 0x80000080, 0x80808000, 0x00000080,
	0x80808000, 0x80808000, 0x80808000, 0x00808000, 0x80808000, 0x80008000, 0x80808000, 0x00008000,
	0x80808000, 0x80800000, 0x80808000, 0x00800000, 0x80808000, 0x80000000, 0x80808000, 0x00000000,
	0x00808000, 0x80808080, 0x00808000, 0x00808080, 0x00808000, 0x80008080, 0x00808000, 0x00008080,
	0x00808000, 0x80800080, 0x00808000, 0x00800080, 0x00808000, 0x80000080, 0x00808000, 0x00000080,
	0x00808000, 0x80808000, 0x00808000, 0x00808000, 0x00808000, 0x80008000, 0x00808000, 0x00008000,
	0x00808000, 0x80800000, 0x00808000, 0x00800000, 0x00808000, 0x80000000, 0x00808000, 0x00000000,
	0x80008000, 0x80808080, 0x80008000, 0x00808080, 0x80008000, 0x80008080, 0x80008000, 0x00008080,
	0x80008000, 0x80800080, 0x80008000, 0x00800080, 0x80008000, 0x80000080, 0x80008000, 0x00000080,
	0x80008000, 0x80808000, 0x80008000, 0x00808000, 0x80008000, 0x80008000, 0x80008000, 0x00008000,
	0x80008000, 0x80800000, 0x80008000, 0x00800000, 0x80008000, 0x80000000, 0x80008000, 0x00000000,
	0x00008000, 0x80808080, 0x00008000, 0x00808080, 0x00008000, 0x80008080, 0x00008000, 0x00008080,
	0x00008000, 0x80800080, 0x00008000, 0x00800080, 0x00008000, 0x80000080, 0x00008000, 0x00000080,
	0x00008000, 0x80808000, 0x00008000, 0x00808000, 0x00008000, 0x80008000, 0x00008000, 0x00008000,
	0x00008000, 0x80800000, 0x00008000, 0x00800000, 0x00008000, 0x80000000, 0x00008000, 0x00000000,
	0x80800000, 0x80808080, 0x80800000, 0x00808080, 0x80800000, 0x80008080, 0x80800000, 0x00008080,
	0x80800000, 0x80800080, 0x80800000, 0x00800080, 0x80800000, 0x80000080, 0x80800000, 0x00000080,
	0x80800000, 0x80808000, 0x80800000, 0x00808000, 0x80800000, 0x80008000, 0x80800000, 0x00008000,
	0x80800000, 0x80800000, 0x80800000, 0x00800000, 0x80800000, 0x80000000, 0x80800000, 0x00000000,
	0x00800000, 0x80808080, 0x00800000, 0x00808080, 0x00800000, 0x80008080, 0x00800000, 0x00008080,
	0x00800000, 0x80800080, 0x00800000, 0x00800080, 0x00800000, 0x80000080, 0x00800000, 0x00000080,
	0x00800000, 0x80808000, 0x00800000, 0x00808000, 0x00800000, 0x80008000, 0x00800000, 0x00008000,
	0x00800000, 0x80800000, 0x00800000, 0x00800000, 0x00800000, 0x80000000, 0x00800000, 0x00000000,
	0x80000000, 0x80808080, 0x80000000, 0x00808080, 0x80000000, 0x80008080, 0x80000000, 0x00008080,
	0x80000000, 0x80800080, 0x80000000, 0x00800080, 0x80000000, 0x80000080, 0x80000000, 0x00000080,
	0x80000000, 0x80808000, 0x80000000, 0x00808000, 0x80000000, 0x80008000, 0x80000000, 0x00008000,
	0x80000000, 0x80800000, 0x80000000, 0x00800000, 0x80000000, 0x80000000, 0x80000000, 0x00000000,
	0x00000000, 0x80808080, 0x00000000, 0x00808080, 0x00000000, 0x80008080, 0x00000000, 0x00008080,
	0x00000000, 0x80800080, 0x00000000, 0x00800080, 0x00000000, 0x80000080, 0x00000000, 0x00000080,
	0x00000000, 0x80808000, 0x00000000, 0x00808000, 0x00000000, 0x80008000, 0x00000000, 0x00008000,
	0x00000000, 0x80800000, 0x00000000, 0x00800000, 0x00000000, 0x80000000, 0x00000000, 0x00000000,
};




static void Turn8x8(uint8_t* dmaBuffer, int iRgbData)
{
	uint32_t tv32[2] = { 0, 0 };
	
	for (int i = 0; i < 8; i++)
	{
		int p = 0;
		
		if (RgbData[i] != 0)
		{
			p = RgbData[i][iRgbData];
		}

		tv32[0] = ((tv32[0] >> 1) & 0x7f7f7f7f) | BitSwapTab[(p << 1) + 0];
		tv32[1] = ((tv32[1] >> 1) & 0x7f7f7f7f) | BitSwapTab[(p << 1) + 1];
	}
	
	#pragma GCC diagnostic push
	#pragma GCC diagnostic ignored "-Wstrict-aliasing"

	uint32_t* db32 = (uint32_t*)&dmaBuffer[0];
		
	#pragma GCC diagnostic pop
	
	db32[0] = tv32[0];
	db32[1] = tv32[1];
}

#endif


// ----------------------------------------------------------------------------------------------


#ifndef WS_RGB2DMA_PORT_FAST

static void Turn8x8(uint8_t* dmaBuffer, uint16_t iRgbData)
{
	__attribute__((aligned(4)))	// Because of word access
	static uint8_t cb[8];
	
	for (int i = 0; i < 8; i++)
	{
		if (RgbData[i] != 0)
		{
			cb[i] = ~RgbData[i][iRgbData];
		}
	}
	
	uint8_t dmaByte;
	
	for (int i = 0; i < 8; i++)
	{
		dmaByte  = cb[0] & 0x80;	dmaByte >>= 1;
		dmaByte |= cb[1] & 0x80;	dmaByte >>= 1;
		dmaByte |= cb[2] & 0x80;	dmaByte >>= 1;
		dmaByte |= cb[3] & 0x80;	dmaByte >>= 1;
		dmaByte |= cb[4] & 0x80;	dmaByte >>= 1;
		dmaByte |= cb[5] & 0x80;	dmaByte >>= 1;
		dmaByte |= cb[6] & 0x80;	dmaByte >>= 1;
		dmaByte |= cb[7] & 0x80;

		*dmaBuffer++ = dmaByte;

		#pragma GCC diagnostic push
		#pragma GCC diagnostic ignored "-Wstrict-aliasing"

		uint32_t* cb32 = (uint32_t*)&cb[0];
		
		#pragma GCC diagnostic pop
		
		cb32[0] = (cb32[0] << 1) & 0xfefefefe;
		cb32[1] = (cb32[1] << 1) & 0xfefefefe;
	}
}

#endif


// ----------------------------------------------------------------------------------------------


int ConvertRgbToDma8Channels(void)
{
	// Check state
	if (!InitOk)	// WsDmaPortInit() not called or error
		return -1;
	
	uint8_t* dmaBuffer = DmaBuffer;

	for (int i = 0; i < (LedCount * 3); i += 3)
	{
		Turn8x8(dmaBuffer +  0, i + 1);	// Green
		Turn8x8(dmaBuffer +  8, i + 0);	// Red
		Turn8x8(dmaBuffer + 16, i + 2);	// Blue

		dmaBuffer += 24;
	}
	
	return 0;
}










